use crate::{
    message_handler::user_commands::UResult,
    models::task::{Task, TaskDetails, TaskMeta},
    regular_tasks::{queued_command::QueuedCommandContext, TaskHandler, TaskRunContext},
    services::{combat::max_health, effect::run_effects},
    static_content::possession_type::UseEffect,
    DResult,
};

use super::{Direction, Exit, ExitTarget, GridCoords, Room, RoomEnterTrigger, SecondaryZoneRecord};
use ansi::ansi;
use async_trait::async_trait;
use chrono::{self, Utc};
use log::warn;
use std::time;

struct EnterERTrigger;
#[async_trait]
impl RoomEnterTrigger for EnterERTrigger {
    async fn handle_enter(
        self: &Self,
        ctx: &mut QueuedCommandContext,
        _room: &Room,
    ) -> UResult<()> {
        ctx.trans
            .upsert_task(&Task {
                meta: TaskMeta {
                    task_code: ctx.item.refstr(),
                    next_scheduled: Utc::now() + chrono::Duration::seconds(60),
                    ..Default::default()
                },
                details: TaskDetails::HospitalERSeePatient {
                    item: ctx.item.refstr(),
                },
            })
            .await?;
        if let Some((sess, _)) = ctx.get_session().await? {
            ctx.trans
                .queue_for_session(
                    &sess,
                    Some(ansi!(
                        "<yellow>The triage nurse says: <reset><bold>\
                            \"Luckily, we are not too busy today - at least not by usual \
                            standards! Take a seat and the doctor will be with you in \
                            about a minute.\"<reset>\n"
                    )),
                )
                .await?;
        }
        Ok(())
    }
}
static ENTER_ER_TRIGGER: EnterERTrigger = EnterERTrigger;

pub struct SeePatientTaskHandler;
#[async_trait]
impl TaskHandler for SeePatientTaskHandler {
    async fn do_task(&self, ctx: &mut TaskRunContext) -> DResult<Option<time::Duration>> {
        let see_who = match ctx.task.details {
            TaskDetails::HospitalERSeePatient { ref item } => item,
            _ => {
                warn!(
                    "Unexpected task dispatched to SeePatientTaskHandler: {:?}",
                    ctx.task
                );
                return Ok(None);
            }
        };
        let (who_type, who_code) = match see_who.split_once("/") {
            None => return Ok(None),
            Some(w) => w,
        };
        let who = match ctx.trans.find_item_by_type_code(who_type, who_code).await? {
            None => return Ok(None),
            Some(w) => w,
        };
        if who.location != "room/general_hospital_waiting_room" {
            return Ok(None);
        }

        if who_type == "player" {
            if let Some((sess, _sess_dat)) = ctx.trans.find_session_for_player(who_code).await? {
                let mut msg: String = ansi!(
                    "<yellow>The doctor says: <reset><bold>\
                     \"Stay still while I take a close look at what might be wrong.\"<reset>\n\
                     The doctor looks up and down your body closely, her gaze searching for any \
                     anomaly.\n"
                )
                .to_owned();
                let hp_gap = max_health(&who) as i64 - who.health as i64;
                let mut skip_heal = false;
                if hp_gap <= 0 {
                    msg += ansi!("<yellow>The doctor says: <reset><bold>\"You're perfectly healthy as far \
                            as I can tell! There's nothing I can do for you here. Eat healthy, try to \
                            maintain a healthy weight, and stay away from radiation, and you'll stay \
                            that way.\"<reset>\n");
                    skip_heal = true;
                } else {
                    msg += ansi!("<yellow>The doctor says: <reset><bold>\"You're injured. Let me stitch that \
                            up, and then just take bed rest. If it stops healing and doesn't feel better, \
                            come right back in to the clinic.\"<reset>\n\
                            The doctor stitches you up, and applies various gels. It hurts, but you \
                            persevere through the pain.\n");
                }
                ctx.trans.queue_for_session(&sess, Some(&msg)).await?;
                if skip_heal {
                    return Ok(None);
                }
            }
        }
        let mut who_mut = (*who).clone();
        run_effects(
            &ctx.trans,
            &vec![
                UseEffect::ChangeTargetHealth {
                    delay_secs: 0,
                    base_effect: 10,
                    skill_multiplier: 0.0,
                    max_effect: 10,
                    message: Box::new(|_item| {
                        (
                            "That feels better".to_owned(),
                            "That feels better".to_owned(),
                        )
                    }),
                },
                UseEffect::ChangeTargetHealth {
                    delay_secs: 10,
                    base_effect: 9,
                    skill_multiplier: 0.0,
                    max_effect: 10,
                    message: Box::new(|_item| {
                        (
                            "That feels better".to_owned(),
                            "That feels better".to_owned(),
                        )
                    }),
                },
                UseEffect::ChangeTargetHealth {
                    delay_secs: 20,
                    base_effect: 8,
                    skill_multiplier: 0.0,
                    max_effect: 10,
                    message: Box::new(|_item| {
                        (
                            "That feels better".to_owned(),
                            "That feels better".to_owned(),
                        )
                    }),
                },
                UseEffect::ChangeTargetHealth {
                    delay_secs: 30,
                    base_effect: 7,
                    skill_multiplier: 0.0,
                    max_effect: 10,
                    message: Box::new(|_item| {
                        (
                            "That feels better".to_owned(),
                            "That feels better".to_owned(),
                        )
                    }),
                },
            ],
            &mut who_mut,
            &who,
            &mut None,
            0.0,
            "bandages",
        )
        .await?;
        ctx.trans.save_item_model(&who_mut).await?;

        Ok(None)
    }
}
pub static SEE_PATIENT_TASK: &'static (dyn TaskHandler + Sync + Send) = &SeePatientTaskHandler;

pub fn room_list() -> Vec<Room> {
    vec!(
        Room {
            zone: "general_hospital",
            secondary_zones: vec!(
                SecondaryZoneRecord {
                    zone: "melbs",
                    short: ansi!("<bgwhite><red>++<reset>"),
                    grid_coords: GridCoords { x: 2, y: 10, z: 0 },
                    caption: Some("General Hospital")
                }
            ),
            code: "general_hospital_waiting_room",
            name: "Emergency Waiting Room",
            short: ansi!("<bgwhite><red>WR<reset>"),
            description: ansi!("A room apparently designed for patients to wait to seen by doctors. \
                                Heavy-duty grey linoleum lines the floors, and even the tops of the walls, \
                                while stainless steel metal strips cover the bottom of the walls. A line on \
                                floor reserves an area of the floor for sick patients to be rushed past on \
                                stretchers. It smells strongly of phenolic cleaners. At the front of the room \
                                a triage nurse assures everyone coming in they will be assessed by doctors \
                                a minute after arriving. Doctors pace the floor treating patients"),
            description_less_explicit: None,
            grid_coords: GridCoords { x: 0, y: 0, z: 0 },
            exits: vec!(
                Exit {
                    direction: Direction::WEST,
                    target: ExitTarget::Custom("room/melbs_kingst_120"),
                    ..Default::default()
                },
            ),
            should_caption: true,
            enter_trigger: Some(&ENTER_ER_TRIGGER),
            ..Default::default()
        },
    )
}
